#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>

void usage(void)
{
  printf("usage: [options] <stage 2 bootloader file> <kernel file>\n");
  printf("options:\n");
  printf("    -ip <ip address to bind to (default 0.0.0.0, any)>\n");
  printf("    -rp <receive port (default 0x6666)>\n");
  printf("    -sp <send port (default 0x6667)>\n");
  printf("    -sip <ip to send to (default 10.255.255.255)>\n");
  printf("        must be the broadcast address of the link (I think)\n");
  exit(1);
}

int main(int n, char **v)
{
  char *stage2_filename = NULL;
  char *kernel_filename = NULL;
  char *filename;
  char buf[1024];
  struct stat sbuf;
  FILE *f;
  int sock;
  struct sockaddr_in addr;
  int rport = 0x6666;
  int sport = 0x6667;
  char *ip = "0.0.0.0";
  char *sip = "10.255.255.255";
  int yes;
  socklen_t alen;
  int i;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-ip")) {if(i>n-2)usage(); ip = v[++i]; continue; }
    if (!strcmp(v[i], "-rp")) {if(i>n-2)usage();rport=atoi(v[++i]);continue;}
    if (!strcmp(v[i], "-sp")) {if(i>n-2)usage();sport=atoi(v[++i]);continue;}
    if (!strcmp(v[i], "-sip")) {if(i>n-2)usage(); sip = v[++i]; continue; }
    if (stage2_filename == NULL) { stage2_filename = v[i]; continue; }
    if (kernel_filename == NULL) { kernel_filename = v[i]; continue; }
    usage();
  }

  if (stage2_filename == NULL || kernel_filename == NULL) usage();

  sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock == -1) { perror("socket"); exit(1); }

  yes = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &yes, sizeof(yes)) == -1)
    { perror("setsockopt"); exit(1); }

  //if (setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE, "eth0", 4) == -1)
  //  { perror("setsockopt"); exit(1); }

  addr.sin_family = AF_INET;
  addr.sin_port = htons(rport);
  addr.sin_addr.s_addr = inet_addr(ip);

  if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1)
    { perror("bind"); exit(1); }

more:
  printf("wait for client...\n");
  alen = sizeof(addr);
  if (recvfrom(sock, buf, 1024, 0, (struct sockaddr *)&addr, &alen) != 1)
    { printf("recvfrom error\n"); exit(1); }
  if (buf[0] == 2)         filename = stage2_filename;
  else if (buf[0] == 0x17) filename = kernel_filename;
  else abort();
  /* we should check what we got here (address, port, data) */
  f = fopen(filename, "r");
  if (f == NULL || stat(filename, &sbuf) == -1)
    { perror(filename); exit(1); }
  printf("serving %s, filesize %ld\n", filename, sbuf.st_size);
  while (sbuf.st_size) {
    size_t l1, l2;
    l1 = 1024; if (l1 > sbuf.st_size) l1 = sbuf.st_size;
    l2 = fread(buf, 1, l1, f);
    if (l1 != l2) { printf("error reading %s\n", filename); exit(1); }
    sbuf.st_size -= l1;
    addr.sin_family = AF_INET;
    addr.sin_port = htons(sport);
    addr.sin_addr.s_addr = inet_addr(sip);
    if (sendto(sock, buf, 1024, 0, (struct sockaddr *)&addr, sizeof(addr)) !=
        1024)
      { perror("sendto"); printf("sendto error\n"); exit(1); }
    alen = sizeof(addr);
    if (recvfrom(sock, buf, 1024, 0, (struct sockaddr *)&addr, &alen) != 1)
      { printf("recvfrom error\n"); exit(1); }
    /* we should check what we got here (address, port, data) */
  }
  addr.sin_family = AF_INET;
  addr.sin_port = htons(sport);
  addr.sin_addr.s_addr = inet_addr(sip);
  if (sendto(sock, buf, 1, 0, (struct sockaddr *)&addr, sizeof(addr)) != 1)
    { printf("sendto error\n"); exit(1); }
  fclose(f);
  printf("...done\n");
  goto more;
  return 0;
}
